/**
   Converter from HTML to DOM ver 1.0

   This script is written by Takaho A. Endo.
   You can freely use, modify, or redistribute this script if you do not alter copyright notification and this comment.

   A class which convert an HTML document to an XML DOM, which is retrieved by XMLHttpRequest.responseText.
   Many tags are allowed as singular such as <P>, <IMG>, <BR>, blah blah blah...
   Therefore you can append tags to a list which consists of tags that are immedently closed using a method 'setSingularTags'. If a document is converted without these items, the DOM tree will be deeply nested in many <IMG> elements.
   Some decorative tags (font, b, i) are neglected because of the simplicity of DOM handling.
   The DOM document has a root element named 'HTMLDOC' and an element 'HTML' will be the first child of the root element.
   This version can interprete only elements, texts and comments. All entities are interpreted as texts.

   Usage:
   var converter = new H2DConverter();
   var document = converter.parse( request.responseText );

   When you would like to user your (probably more simple) parser, you can alter parsing object using H2DConverter.setParser method. The parser must have following methods:
   startElement( qName, attributes )
   endElement( qName )
   characters( data )
   comment( data ).

   You can use useful functions getElementAt and getInsideText in following codes (see comments at the header of each function).

   ̃XNvgTakaho A. EndoɌA܂B
   ̃XNvg͎RɎgpAρAĔzz\łB쌠\Ƃ̃Rg͕ύXȂƂv܂B

   XMLHttRequest.responseTextŎ擾eLXgXML DOMfɕϊNX
   HTML͏I^OȂĂLȃ^O邪AsetSingularTagsł^Oꂽ炷ɏI邱ƂłBɂDOMimgvf̐[ʒuɖ̂hƂłB
   ̃^ODOM̒P̂߂ɖB
   DOMhLg̓[gƂ'HTMLDOC'Gg̒ɓB͎擾HTMLHTMLvf̊ODTDeLXgȂǂĂ邱Ƃ邩łB
   ݂̃o[Wł̓GeBeB[ƃeLXgʂĂȂB

   Usage:
   var converter = new H2DConverter();
   var document = converter.parse( request.responseText );

   ܂Ǝ̃p[T[ɂăVvHTML߂sꍇsetParser\bhŃp[T[ւ邱ƂłB̃p[T[
   startElement( qName, attributes )
   endElement( qName )
   characters( data )
   comment( data )
   ̂S̃\bhAdocumenttB[hĂȂ΂ȂȂ
   
   ̂ق֗Ȋ֐(getElementAt, getInsideTexts)ȉ̃R[hɂ邪Ae֐̐͂ꂼ̃RgɁB
*/

function H2DConverter() {
    this.parser = new _HTMLHandler();
    this.attrRegex = new RegExp( "([a-zA-Z0-9_\\-]+)(\\s*=\\s*)?(('[^']*?')|(\"[^\"]*?\")|([^'\"\\s]*))", "gm" );
    this.nullArrtibute = new Array();
    // tags that can appear without end tags.
    this.singleTags = [ "BR", "META", "IMG", "HR", "INPUT", "PARAM" ];
    // tags that may not have nested element of the same tags
    this.nestInhibitedTags = [ "LI", "TR", "TD", "OPTION" ];
    // tags which are ignored
    this.ignoringTags = [ "FONT", "FONTSIZE", "B", "I" ];
}

/**
   Create attribute from a text in an element tag.
 */
function H2DConverter__createAttribute( text ) {
    this.attrRegex.lastIndex = 0;
    var am = this.attrRegex.exec( text );
    var at = null;
        
    if ( am ) {
        at = new Array();
        while ( am ) {
            var label = am[ 1 ];
            var value;
            if ( am[ 2 ] ) {
                if ( am[ 4 ] != null || am[ 5 ] != null ) {
                    value = am[ 3 ].substring( 1, am[ 3 ].length - 1 );
                } else if ( am[ 6 ] ) {
                    value = am[ 6 ];
                } else {
                    value = "";
                }
                if ( label != null && value != null ) {
                    at.push( new Array( label, value ) );
                }
            }
            am = this.attrRegex.exec( text );
        }
        return at;
    } else {
        return this.nullAttribute;
    }
}

/**
   Parse HTML document and call HTML handler.
 */
function H2DConverter__parse( text ) {
    var _HTML_elementRegex = new RegExp( "(<([a-zA-Z][0-9\\-_A-Za-z]*)\\s*([^>]*)>)|(</([^>]+)\\s*>)", "img" );
    var _HTML_spaceRegex = new RegExp( "[\\s\\W]+" );
    var _HTML_commentRegex = new RegExp( "\\s*<!\-\-(.*?)\-\->\\s*", "gm" );
    var matches = _HTML_elementRegex.exec( text );
    var pos = 0;

    while ( matches ) {
        var start = matches[ 2 ];
        var end   = matches[ 4 ];

        var pos_match = _HTML_elementRegex.lastIndex - matches[ 0 ].length;
        // strip the text and append Text if the text is not null
        var p0;
        var p1;
        for ( p0 = pos; p0 < pos_match; p0++ ) {
            var ct = text.charAt( p0 );
            if ( ct != "\n" && ct != " " ) {
                break;
            }
        }
        for ( p1 = pos_match - 1; p1 >= pos; p1 -- ) {
            var ct = text.charAt( p1 );
            if ( ct != "\n" && ct != " " ) {
                p1 ++;
                break;
            }
        }

        // separate texts and comments and append to parent node
        if ( p1 - p0 > 0 ) {
            // remove comment and apply it
            _HTML_commentRegex.lastIndex = 0;
            var chrs = text.substring( p0, p1 );
            var cp0 = 0;
            var cp1 = chrs.length;
            var cmmatches = _HTML_commentRegex.exec( chrs );
            while ( cmmatches ) {
                var cp1 = _HTML_commentRegex.lastIndex - cmmatches[ 0 ].length;
                if ( cp1 - cp0 > 0 ) {
                    this.parser.characters( chrs.substring( cp0, cp1 ) );
                }
                cp0 = _HTML_commentRegex.lastIndex;
                this.parser.comment( cmmatches[ 1 ] );
                cmmatches = _HTML_commentRegex.exec( chrs );
            }
            if ( cp0 < chrs.length ) {
                this.parser.characters( chrs.substring( cp0, chrs.lenth ) );
            }
        }
 
        if ( start ) {
            start = start.toUpperCase();
            // determine the tag can be ignored
            var flag_ignore = false;
            for ( var j = 0; j < this.ignoringTags.length; j++ ) {
                if ( this.ignoringTags[ j ] == start ) {
                    flag_ignore = true;
                    break;
                }
            }

            // check nesting restriction
            var flag_nest = false;
            if ( flag_ignore == false && this.parser.currentElement != null 
                 && this.parser.currentElement.tagName == start ) {
                for ( var j = 0; j < this.nestInhibitedTags.length; j++ ) {
                    if ( this.nestInhibitedTags[ j ] == start ) {
                        flag_nest = true;
                        this.parser.endElement( start );
                        break;
                    }
                }
            }

            // valid tag
            if ( flag_ignore == false ) {
                var attrs = this.createAttribute( matches[ 3 ] );
                this.parser.startElement( start, attrs );
                // if the tagName is one of singular, endElement
                // is called soon after startElement
                try {
                    for ( var j = 0; j < this.singleTags.length; j++ ) {
                        if ( this.singleTags[ j ] == start ) {
                            this.parser.endElement( start );
                            break;
                        }
                    }
                } catch ( e ) {
                    //throw e;
                    //debug( "multiple document, exit" );
                    break;
                }
            }
        } else if ( end ) {
            this.parser.endElement( matches[ 5 ].toUpperCase() );
        }

        pos = pos_match + matches[ 0 ].length;
        matches = _HTML_elementRegex.exec( text );
    }
    return this.parser.document;
}

// definitions of methods
// \bh̒`
H2DConverter.prototype.parse           = H2DConverter__parse;
H2DConverter.prototype.createAttribute = H2DConverter__createAttribute;
H2DConverter.prototype.setHandler      = function( handler ) { this.parser = handler;  };
H2DConverter.prototype.setSingularTags = function( tags ) { this.singleTags = tags; };
H2DConverter.prototype.setNestlessTags = function( tags ) { this.nestInhibitedTags = tags; };
H2DConverter.prototype.setIgnoringTags = function( tags ) { this.ignoringTags = tags; };
H2DConverter.prototype.setParser       = function( parser ) { this.parser = parser; };


// H2DConverterĂяop[T
// SAX handler of HTML document
function _HTMLHandler() {
    this.document = document.implementation.createDocument( "", "", null );
    this.rootElement     = this.document.createElement( "HTMLDOC" );
    this.currentElement  = this.rootElement;
    this.document.appendChild( this.rootElement );
}

// begin element
function _HTMLHandler__startElement( qName, attrs ) {
    var elem = this.document.createElement( qName );
    if ( attrs ) {
        for ( var i = 0; i < attrs.length; i++ ) {
            elem.setAttribute( attrs[ i ][ 0 ], attrs[ i ][ 1 ] );
        }
    }
    if ( this.currentElement ) {
        this.currentElement.appendChild( elem );
    } else {
        if ( this.rootElement ) {
            throw "two documents";
        }
        this.rootElement = elem;
        this.document.appendChild( elem );
    }
    this.currentElement = elem;
}

// end element
function _HTMLHandler__endElement( qName ) {
    if ( qName == this.currentElement.tagName ) {
        if ( this.currentElement ) {
            this.currentElement = this.currentElement.parentNode;
        } else { // no element, exit parsing
            this.currentElement = null;
        }
        return;
    } else {
        // mismatch
        var elem = this.currentElement.parentNode;
        while ( elem ) {
            if ( elem.tagName == qName ) {
                this.currentElement = elem.parentNode;
                return;
            }
            elem = elem.parentNode;
        }
    }
}

// text
function _HTMLHandler__characters( text ) {
    if ( text == null || text.length == 0 ) return;
    var tx = this.document.createTextNode( text );
    if ( this.currentElement ) {
        this.currentElement.appendChild( tx );
    }
}

function _HTMLHandler__comment( commenttext ) {
    if ( commenttext == null ) return;
    var cm = this.document.createComment( commenttext );
    if ( this.currentElement ) {
        this.currentElement.appendChild( cm );
    }
}

_HTMLHandler.prototype.startElement = _HTMLHandler__startElement;
_HTMLHandler.prototype.endElement = _HTMLHandler__endElement;
_HTMLHandler.prototype.characters = _HTMLHandler__characters;
_HTMLHandler.prototype.comment = _HTMLHandler__comment;

/**
   DOM̃eLXg悭o߂̊֐B
   var insideTexts = getInsideTexts( element );
   Ƃ邱Ƃłvf̃eLXg𒊏ołB
   Ⴄvf̃eLXgƂ̃Zp[^wł̂ŁAႦ
   getInsideTexts( element, ',' ).split( ',' )
   Ƃ邱ƂŔz`ŃeLXgǂݏoƂłB

*/
function getInsideTexts( elem, separator, path ) {
    if ( path ) {
        elem = getElementAt( elem, path );
    }
    if ( elem == null ) return;
    if ( separator == undefined ) separator = "";
    return _getTextAt( elem, separator );
}


function _getTextAt( elem, separator ) {
    var children = elem.childNodes;
    var output = "";
    for ( var i = 0; i < children.length; i++ ) {
        var child = children.item( i );
        if ( child.nodeType == 3 ) { // text
            if ( output == null ) output = child.data;
            else output += separator + child.data;
        } else if ( child.nodeType == 1 ) {
            if ( output == null ) output = _getTextAt( child, separator );
            else output += separator + _getTextAt( child, separator );
        }
    }
    return output;
}

/** 
    get Element node at a path.
    A path is given as "my/element/path" for <root><my><element><path></path></element></my></root>

    vf̒ɂElement擾BȂnullԂB
    DTD̂ȂHTMLł͂vf̒ɑ̗vf邩P邩킩Ȃ̂ŃANZX@ƂE2X̂悤ȕ@Ƃ̂͐ȂB
    ̊֐
    var element = getElementAt( parent, "/A/B,2/C,3" );
    ƃ^OƂ̒̏ԂJ}ŐڑBPł΍ŏ̗vfłB
 */
function getElementAt( elem, path ) {
    var names = path.toUpperCase().split( "/" );
    var children = elem.childNodes;
    return _getElementAt( elem, names, 0 );
}

// recursive function called by getElementAt
function _getElementAt( elem, names, num ) {
    var items = names[ num ].split( "," );
    var cn = 1;
    if ( items.length > 1 ) {
        cn = parseInt( items[ 1 ] );
    }
    var children = elem.childNodes;
    var last = ( names.length - 1 == num );
    for ( var i = 0; i < children.length; i++ ) {
        var child = children.item( i );
        if ( child.nodeType == 1 ) {
            if ( child.tagName == items[ 0 ] ) {
                cn --;
                if ( cn <= 0 ) {
                    if ( last ) return child;
                    return _getElementAt( child, names, num + 1 );
                }
            }
        }
    }
    return null;
}


/**
   Output a DOM tree with a callback function.
*/
function outputDOMText( elem, callback ) {
    if ( callback == null ) {
        return;
    }
    _outputDOMText( elem, "", elem, callback );
}

function _outputDOMText( elem, indent, parent, callback ) {
    var output = indent + "<" + elem.tagName;
    var attrs = elem.attributes;
    for ( var i = 0; i < attrs.length; i++ ) {
        var label = attrs.item( i ).name;
        var value = attrs.item( i ).value;
        output += " " + label + "=\"" + value + "\"";
    }
    var _children = elem.childNodes;
    if ( _children.length > 0 ) {
        output += "> <!-- " + _getPathFromParent( elem, parent ) + " -->";
        callback( output );
        var indent_next = indent + "  ";
        for ( var i = 0; i < _children.length; i++ ) {
            var child = _children.item( i );
            if ( child.nodeType == 1 ) {
                _outputDOMText( child, indent_next, parent, callback );
            } else if ( child.nodeType == 3 ) {
                callback( indent_next + child.data );
            } else if ( child.nodeType == 8 ) { // comment
                callback( indent_next + "<!-- " + child.data + " -->" );
            }
        }
        
        callback( indent + "</" + elem.tagName + ">" );
    } else {
        output += "/>  <!-- " + _getPathFromParent( elem, parent ) + " -->";
        callback( output );
    }
}

function convertDOMToText( elem ) {
    return _convertDOMToText( elem, "" );
}

function _convertDOMToText( elem, indent ) {
    var idtext = indent + " ";
    var output = idtext + "<" + elem.tagName;
    var attrs = elem.attributes;
    for ( var i = 0; i < attrs.length; i++ ) {
        var label = attrs.item( i ).name;
        var value = attrs.item( i ).value;
        output += " " + label + "=\"" + value + "\"";
    }

    if ( elem.childNodes.length == 0 ) {
        output += "/>";
    } else {
        output += ">";
        
        var _children = elem.childNodes;
        for ( var i = 0; i < _children.length; i++ ) {
            var child = _children.item( i );
            if ( child.nodeType == 1 ) {
                output += _convertDOMToText( child, idtext );
            } else if ( child.nodeType == 3 ) {
                output += idtext + child.data + "\n";
            } else if ( child.nodeType == 8 ) { // comment
                output += idtext + "<<<!-- " + child.data + " -->>>\n";
            }
        }
        output += idtext + "</" + elem.tagName + ">\n";
    }
    return output;
}

function _getPathFromParent( elem, parent ) {
    var path = elem.tagName;
    if ( elem == parent ) return path;
    var el = elem.parentNode;
    while ( el ) {
        if ( el.nodeType == undefined ) break;
        if ( el.nodeType != 1 ) {
            return path;
        }
        path = el.tagName + "/" + path;
        if ( el == parent ) {
            return path;
        }
        el = el.parentNode;
    }
    return path;
}

